#!/usr/bin/env python3

import math
import numpy
from pyquats.quats import Quat
from pyquats.qtools import random_move_quat
from pyquats.qtools import random_move_Xwall
from pyquats.qtools import random_move_Ywall
from pyquats.qtools import random_move_Zwall

#unit_quat = Quat.from_x_rotation(math.pi)
#unit_quat = Quat.from_y_rotation(math.pi / 2.)
#unit_quat = Quat.from_x_rotation(math.pi * 0.5) * Quat.from_y_rotation(math.pi * 0.5) # n||x, l||y
#unit_quat = Quat.from_y_rotation(math.pi * 0.5) * Quat.from_x_rotation(math.pi * 0.5) # n||-y, l||-z
#unit_quat = Quat.from_z_rotation(math.pi / 2.)
#unit_quat = Quat.from_z_rotation(math.pi / 6.)
#unit_quat = Quat.from_z_rotation(math.pi / 6.) * Quat.from_y_rotation(math.pi / 2.)
#unit_quat = Quat.from_z_rotation(math.pi / 4.) * Quat.from_y_rotation(math.pi / 2.)
#unit_quat = random_move_quat(0.5)
unit_quat = random_move_Zwall(0.5)

q0, q1, q2, q3 = unit_quat.q

l_x = 2.0 * (q0 * q0 + q1 * q1) - 1.0
l_y = 2.0 * (q0 * q3 + q1 * q2)
l_z = 2.0 * (q1 * q3 - q0 * q2)
m_x = 2.0 * (q1 * q2 - q0 * q3)
m_y = 2.0 * (q0 * q0 + q2 * q2) - 1.0
m_z = 2.0 * (q0 * q1 + q2 * q3)
n_x = 2.0 * (q0 * q2 + q1 * q3)
n_y = 2.0 * (q2 * q3 - q0 * q1)
n_z = 2.0 * (q0 * q0 + q3 * q3) - 1.0

print ( "unit_quat", repr(unit_quat) )
print ( "vector l", l_x, l_y, l_z )
print ( "vector m", m_x, m_y, m_z )
print ( "vector n", n_x, n_y, n_z )

# Sprawdzam dzialania na macierzach.
Qll = numpy.zeros((3,3), dtype=float)
Qmm = numpy.zeros((3,3), dtype=float)
Qnn = numpy.zeros((3,3), dtype=float)

Qll[0, 0] += 1.5 * l_x * l_x - 0.5
Qll[0, 1] += 1.5 * l_x * l_y
Qll[0, 2] += 1.5 * l_x * l_z
Qll[1, 1] += 1.5 * l_y * l_y - 0.5
Qll[1, 2] += 1.5 * l_y * l_z
Qll[2, 2] += 1.5 * l_z * l_z - 0.5
Qll[1, 0] = Qll[0, 1]
Qll[2, 0] = Qll[0, 2]
Qll[2, 1] = Qll[1, 2]

Qmm[0, 0] += 1.5 * m_x * m_x - 0.5
Qmm[0, 1] += 1.5 * m_x * m_y
Qmm[0, 2] += 1.5 * m_x * m_z
Qmm[1, 1] += 1.5 * m_y * m_y - 0.5
Qmm[1, 2] += 1.5 * m_y * m_z
Qmm[2, 2] += 1.5 * m_z * m_z - 0.5
Qmm[1, 0] = Qmm[0, 1]
Qmm[2, 0] = Qmm[0, 2]
Qmm[2, 1] = Qmm[1, 2]

Qnn[0, 0] += 1.5 * n_x * n_x - 0.5
Qnn[0, 1] += 1.5 * n_x * n_y
Qnn[0, 2] += 1.5 * n_x * n_z
Qnn[1, 1] += 1.5 * n_y * n_y - 0.5
Qnn[1, 2] += 1.5 * n_y * n_z
Qnn[2, 2] += 1.5 * n_z * n_z - 0.5
Qnn[1, 0] = Qnn[0, 1]
Qnn[2, 0] = Qnn[0, 2]
Qnn[2, 1] = Qnn[1, 2]

print ( "Qll", Qll )
print ( "Qmm", Qmm )
print ( "Qnn", Qnn )

print ( "Manual settings for Q" )

# Konfiguracja idealna na scianie prostopadlej do Z.
Qll_ideal = numpy.array([[-0.5, 0, 0], [0, -0.5, 0], [0, 0, 1]], dtype=float)
Qmm_ideal = numpy.array([[-0.5, 0, 0], [0, 1, 0], [0, 0, -0.5]], dtype=float)
Qnn_ideal = numpy.array([[1, 0, 0], [0, -0.5, 0], [0, 0, -0.5]], dtype=float)

print ( "Qll_ideal", Qll_ideal )
print ( "Qmm_ideal", Qmm_ideal )
print ( "Qnn_ideal", Qnn_ideal )

Qll = (Qll + Qll_ideal) / 2.
Qmm = (Qmm + Qmm_ideal) / 2.
Qnn = (Qnn + Qnn_ideal) / 2.

print ( "Qll", Qll )
print ( "Qmm", Qmm )
print ( "Qnn", Qnn )

# Wektory wlasne beda ulozone kolumnami!
# The column vec[:,i] is the eigenvector corresponding to the eigenvalue w[i].
w, vec =  numpy.linalg.eig(Qnn)
#print (w, vec)
print ("eigenvalues", w)
print ("unit eigenvectors (in columns)")
print (vec)

print ( "Diagonalizacja B^T Qll B ..." )
print ( numpy.dot(numpy.transpose(vec), numpy.dot(Qll,vec)) )
print ( "Diagonalizacja B^T Qmm B ..." )
print ( numpy.dot(numpy.transpose(vec), numpy.dot(Qmm,vec)) )
print ( "Diagonalizacja B^T Qnn B ..." )
print ( numpy.dot(numpy.transpose(vec), numpy.dot(Qnn,vec)) )

# EOF



#!/usr/bin/env python3

import math
import random
from pyquats.quats import Quat
#from pyquats.numpyquats import Quat

# Rotates a vector with respect to a quat.
# The vector is from R^3, the result is from R^3, rot_quat is a unit quat.
def rotate1(vector, unit_quat):
    """Return the rotated vector (rot_quat used)."""
    # From a vector to a quaternion.
    vec_quat = Quat(0, vector[0], vector[1], vector[2])
    # Rotate a vector.
    vec_quat = unit_quat * vec_quat * (~unit_quat)
    # Return a vector from R^3.
    return vec_quat.q[1:]

# Rotates a vector angle 'radians' with respect to axis,
# both are vectors but it uses quaternions.
def rotate2(vector, axis, angle):
    """Return the rotated vector (axis and angle used)."""
    # Make a unit quat.
    unit_quat = Quat.rot_quat(axis, angle)
    return rotate1(vector, unit_quat)

def rotate3(vector, phi, theta, psi):
    """Return the rotated vector (Euler angles used)."""
    unit_quat = Quat.from_eulers(phi, theta, psi)
    return rotate1(vector, unit_quat)

def random_quat_uniax():
    """Return a random rotation quat for uniaxial molecules."""
    phi = random.uniform(0, 2*math.pi)
    ct = random.uniform(-1, 1)
    theta = math.acos(ct)   # mozna tez -ct
    quat = Quat.from_eulers(phi, theta, 0)
    return quat

def random_quat_biax():
    """Return a random rotation quat for biaxial molecules."""
    phi = random.uniform(0, 2*math.pi)
    ct = random.uniform(-1, 1)
    theta = math.acos(ct)   # mozna tez -ct
    psi = random.uniform(0, 2*math.pi)
    quat = Quat.from_eulers(phi, theta, psi)
    return quat

def random_quat_Xwall():
    """Return a random rotation quat for biaxial molecules at a X wall."""
    alpha = random.uniform(0, 2*math.pi)
    beta = random.choice((0, math.pi))
    #print("beta {}".format(beta))
    quat = Quat.from_x_rotation(alpha)
    quat *= Quat.from_z_rotation(beta)
    return quat

def random_quat_Ywall():
    """Return a random rotation quat for biaxial molecules at a Y wall."""
    alpha = random.uniform(0, 2*math.pi)
    beta = math.pi * random.choice((0.5, 1.5))
    #print("beta {}".format(beta))
    quat = Quat.from_y_rotation(alpha)
    quat *= Quat.from_z_rotation(beta)
    return quat

def random_quat_Zwall():
    """Return a random rotation quat for biaxial molecules at a Z wall."""
    alpha = random.uniform(0, 2*math.pi)
    beta = math.pi * random.choice((0.5, 1.5))
    #print("beta {}".format(beta))
    quat = Quat.from_z_rotation(alpha)
    quat *= Quat.from_y_rotation(beta)
    return quat

def random_pair():
    """Return a random pair (s1, s2), where s1^2 + s2^2 < 1."""
    while True:
        s1 = random.uniform(-1, 1)
        s2 = random.uniform(-1, 1)
        if s1 * s1 + s2 * s2 < 1:
            break
    return s1, s2

def random_unit_quat():
    """Return a random unit quat (Marsaglia, 1972)."""
    s1, s2 = random_pair()
    s3, s4 = random_pair()
    S1 = s1 * s1 + s2 * s2
    S2 = s3 * s3 + s4 * s4
    a = math.sqrt((1.0 - S1) / S2)
    return Quat(s1, s2, s3 * a, s4 * a)

def random_move_quat(ksi=1.0):
    """Return a Monte Carlo move (Barker, Watts, 1969)."""
    assert 0.0 <= ksi <= 1.0
    s1, s2 = random_pair()
    S = s1 * s1 + s2 * s2
    q1 = 2 * s1 * math.sqrt(1-S)
    q2 = 2 * s2 * math.sqrt(1-S)
    q3 = 1 - 2 * S
    # [q1, q2, q3] is now a unit vector.
    angle = random.uniform(-1, 1) * math.pi * 0.5 * 0.5 * ksi
    # math.pi * 0.5 gives all rotations,
    # 0.5 is for quat cos(angle/2), sin(angle/2).
    q0 = math.cos(angle)
    sinus = math.sin(angle)
    q1 *= sinus
    q2 *= sinus
    q3 *= sinus
    return Quat(q0, q1, q2, q3)

def random_move_Xwall(ksi=1.0):
    """Return a Monte Carlo move at a X wall."""
    # For biaxial molecules ksi=0.5 gives the same as ksi=1.0.
    # ksi=0.25 is better for low temperatures.
    alpha = random.uniform(-1, 1) * math.pi * ksi
    quat = Quat.from_x_rotation(alpha)
    return quat

def random_move_Ywall(ksi=1.0):
    """Return a Monte Carlo move at a Y wall."""
    # For biaxial molecules ksi=0.5 gives the same as ksi=1.0.
    # ksi=0.25 is better for low temperatures.
    alpha = random.uniform(-1, 1) * math.pi * ksi
    quat = Quat.from_y_rotation(alpha)
    return quat

def random_move_Zwall(ksi=1.0):
    """Return a Monte Carlo move at a Z wall."""
    # For biaxial molecules ksi=0.5 gives the same as ksi=1.0.
    # ksi=0.25 is better for low temperatures.
    alpha = random.uniform(-1, 1) * math.pi * ksi
    quat = Quat.from_z_rotation(alpha)
    return quat

# EOF



#!/usr/bin/env python3

import math
import numpy
from pyquats.quats import Quat

unit_quat = Quat(1) # R_z(0)
#unit_quat = Quat.from_z_rotation(0) # R_z(0)
#unit_quat = Quat.from_z_rotation(math.pi) # R_z(pi)
#unit_quat = Quat.from_y_rotation(math.pi) # R_y(pi)
#unit_quat = Quat.from_x_rotation(math.pi) # R_x(pi)
#unit_quat = Quat.from_z_rotation(math.pi * 0.5) # R_z(pi/2)
#unit_quat = Quat.from_y_rotation(math.pi * 0.5) # R_y(pi/2)
#unit_quat = Quat.from_x_rotation(math.pi * 0.5) # R_x(pi/2)
#unit_quat = Quat(0.5, 0.5, 0.5, 0.5)   # os [1,1,1], kat 120 stopni
#unit_quat = Quat.from_z_rotation(math.pi * 0.25) # R_z(pi/4)
#unit_quat = Quat.from_z_rotation(math.pi / 6.) # R_z(pi/6)

q0, q1, q2, q3 = unit_quat.q

l_x = 2.0 * (q0 * q0 + q1 * q1) - 1.0
l_y = 2.0 * (q0 * q3 + q1 * q2)
l_z = 2.0 * (q1 * q3 - q0 * q2)
m_x = 2.0 * (q1 * q2 - q0 * q3)
m_y = 2.0 * (q0 * q0 + q2 * q2) - 1.0
m_z = 2.0 * (q0 * q1 + q2 * q3)
n_x = 2.0 * (q0 * q2 + q1 * q3)
n_y = 2.0 * (q2 * q3 - q0 * q1)
n_z = 2.0 * (q0 * q0 + q3 * q3) - 1.0

print ( "unit_quat", repr(unit_quat) )
print ( "vector l", l_x, l_y, l_z )
print ( "vector m", m_x, m_y, m_z )
print ( "vector n", n_x, n_y, n_z )

# Sprawdzam dzialania na macierzach.
Qll = numpy.zeros((3,3), dtype=float)
Qmm = numpy.zeros((3,3), dtype=float)
Qnn = numpy.zeros((3,3), dtype=float)

Qll[0, 0] += 1.5 * l_x * l_x - 0.5
Qll[0, 1] += 1.5 * l_x * l_y
Qll[0, 2] += 1.5 * l_x * l_z
Qll[1, 1] += 1.5 * l_y * l_y - 0.5
Qll[1, 2] += 1.5 * l_y * l_z
Qll[2, 2] += 1.5 * l_z * l_z - 0.5
Qll[1, 0] = Qll[0, 1]
Qll[2, 0] = Qll[0, 2]
Qll[2, 1] = Qll[1, 2]

Qmm[0, 0] += 1.5 * m_x * m_x - 0.5
Qmm[0, 1] += 1.5 * m_x * m_y
Qmm[0, 2] += 1.5 * m_x * m_z
Qmm[1, 1] += 1.5 * m_y * m_y - 0.5
Qmm[1, 2] += 1.5 * m_y * m_z
Qmm[2, 2] += 1.5 * m_z * m_z - 0.5
Qmm[1, 0] = Qmm[0, 1]
Qmm[2, 0] = Qmm[0, 2]
Qmm[2, 1] = Qmm[1, 2]

Qnn[0, 0] += 1.5 * n_x * n_x - 0.5
Qnn[0, 1] += 1.5 * n_x * n_y
Qnn[0, 2] += 1.5 * n_x * n_z
Qnn[1, 1] += 1.5 * n_y * n_y - 0.5
Qnn[1, 2] += 1.5 * n_y * n_z
Qnn[2, 2] += 1.5 * n_z * n_z - 0.5
Qnn[1, 0] = Qnn[0, 1]
Qnn[2, 0] = Qnn[0, 2]
Qnn[2, 1] = Qnn[1, 2]

print ( "Qll", Qll )
print ( "Qmm", Qmm )
print ( "Qnn", Qnn )

print ( "Manual settings for Q" )

# Konfiguracja idealna R_z(0).
Qll_ideal = numpy.array([[1, 0, 0], [0, -0.5, 0], [0, 0, -0.5]], dtype=float)
Qmm_ideal = numpy.array([[-0.5, 0, 0], [0, 1, 0], [0, 0, -0.5]], dtype=float)
Qnn_ideal = numpy.array([[-0.5, 0, 0], [0, -0.5, 0], [0, 0, 1]], dtype=float)

# for R_z(pi/4)
# Do tensorow idealnych dodaje tensory obrocone R_z(pi/4)
# i wynik dziele przez 2.
#Qll = numpy.array([[5./8., 3./8., 0], [3./8., -1./8., 0], [0, 0, -0.5]], dtype=float)
#Qmm = numpy.array([[-1/8., -3/8., 0], [-3/8., 5/8., 0], [0, 0, -0.5]], dtype=float)
#Qnn = numpy.array([[-0.5, 0, 0], [0, -0.5, 0], [0, 0, 1]], dtype=float)

# for R_z(pi/6)
# Do tensorow idealnych dodaje tensory obrocone R_z(pi/6)
# i wynik dziele przez 2.
a = 3*math.sqrt(3)/16.
Qll = numpy.array([[13./16., a, 0], [a, -5./16., 0], [0, 0, -0.5]], dtype=float)
Qmm = numpy.array([[-5/16., -a, 0], [-a, 13/16., 0], [0, 0, -0.5]], dtype=float)
Qnn = numpy.array([[-0.5, 0, 0], [0, -0.5, 0], [0, 0, 1]], dtype=float)

print ( "Qll", Qll )
print ( "Qmm", Qmm )
print ( "Qnn", Qnn )

print ( "eigenvalues ll", numpy.linalg.eigvals(Qll) )
print ( "eigenvalues mm", numpy.linalg.eigvals(Qmm) )
print ( "eigenvalues nn", numpy.linalg.eigvals(Qnn) )

print ( "eig ll", numpy.linalg.eig(Qll) )
print ( "eig mm", numpy.linalg.eig(Qmm) )
print ( "eig nn", numpy.linalg.eig(Qnn) )

# EOF



#!/usr/bin/env python3

import unittest
import math
from pyquats.quats import Quat
from pyquats.qtools import rotate1
from pyquats.qtools import rotate2
from pyquats.qtools import rotate3
from pyquats.qtools import random_quat_uniax
from pyquats.qtools import random_quat_biax
from pyquats.qtools import random_unit_quat
from pyquats.qtools import random_move_quat
from pyquats.qtools import random_quat_Xwall
from pyquats.qtools import random_quat_Ywall
from pyquats.qtools import random_quat_Zwall
from pyquats.qtools import random_move_Xwall
from pyquats.qtools import random_move_Ywall
from pyquats.qtools import random_move_Zwall


class TestRotations(unittest.TestCase):

    def setUp(self):
        self.rotX180 = Quat(0, 1, 0, 0)
        self.rotY180 = Quat(0, 0, 1, 0)
        self.rotZ180 = Quat(0, 0, 0, 1)
        p2 = math.sqrt(0.5)
        p3 = math.sqrt(1.0 / 3.0)
        self.rotX90 = Quat(p2, p2, 0, 0)
        self.rotY90 = Quat(p2, 0, p2, 0)
        self.rotZ90 = Quat(p2, 0, 0, p2)
        # obrot o kat 120 stopni wokol [0,1,1,1]
        self.rot120 = Quat(0.5, 0.5, 0.5, 0.5)
        self.vec1 = [p3, p3, p3]

    def test_rot_quat(self):
        # obroty o 180 sa przemienne, ale daja inne quat
        self.assertAlmostEqual(self.rotX180 * self.rotY180, Quat(0, 0, 0, 1))
        self.assertAlmostEqual(self.rotY180 * self.rotX180, Quat(0, 0, 0, -1))
        # 90+90=180
        # assertAlmostEqual potrzebuje metody __abs__ do dzialania
        self.assertAlmostEqual(self.rotX90 * self.rotX90, self.rotX180)
        self.assertAlmostEqual(self.rotY90 * self.rotY90, self.rotY180)
        self.assertAlmostEqual(self.rotZ90 * self.rotZ90, self.rotZ180)
        # obroty o 90
        self.assertAlmostEqual(self.rotX90 * self.rotY90, self.rot120)
        self.assertAlmostEqual(self.rotY90 * self.rotZ90, self.rot120)
        self.assertAlmostEqual(self.rotZ90 * self.rotX90, self.rot120)
        self.assertAlmostEqual(self.rotY90 * self.rotX90, Quat(0.5, 0.5, 0.5, -0.5))
        self.assertAlmostEqual(self.rotZ90 * self.rotY90, Quat(0.5, -0.5, 0.5, 0.5))
        self.assertAlmostEqual(self.rotX90 * self.rotZ90, Quat(0.5, 0.5, -0.5, 0.5))
        self.assertAlmostEqual(self.rotX90, Quat.rot_quat([1, 0, 0], math.pi * 0.5))
        self.assertAlmostEqual(self.rotY90, Quat.rot_quat([0, 1, 0], math.pi * 0.5))
        self.assertAlmostEqual(self.rotZ90, Quat.rot_quat([0, 0, 1], math.pi * 0.5))
        self.assertAlmostEqual(self.rot120, Quat.rot_quat(self.vec1, math.pi * 2.0 / 3.0))

    def test_rot_vectors(self):
        # Listy nie mozna odejmowac i obliczac abs(L).
        # Potrzebna klasa Vector.
        pairs = []
        pairs.append((rotate1([10, 0, 0], self.rotZ90), [0, 10, 0]))
        pairs.append((rotate2([10, 0, 0], [0, 0, 1], math.pi * 0.5), [0, 10, 0]))
        pairs.append((rotate1([0,10, 0], self.rot120), [0, 0, 10]))
        pairs.append((rotate2([0, 10, 0], self.vec1, math.pi * 2.0 / 3.0), [0, 0, 10]))
        pairs.append((rotate3([0, 0, 10], 0, math.pi * 0.5, 0), [10, 0, 0]))
        pairs.append((rotate3([10, 0, 0], math.pi * 0.5, 0, 0), [0, 10, 0]))
        pairs.append((rotate3([10, 0, 0], math.pi * 0.5, 0, 0), [0, 10, 0]))
        # Po kolei porownuje wyniki po wspolrzednych.
        for left, right in pairs:
            self.assertAlmostEqual(left[0], right[0])
            self.assertAlmostEqual(left[1], right[1])
            self.assertAlmostEqual(left[2], right[2])

    def test_random_quat(self):
        uniax = random_quat_uniax()
        biax = random_quat_biax()
        unit = random_unit_quat()
        move = random_move_quat(0.5)
        self.assertAlmostEqual(abs(uniax), 1)
        self.assertAlmostEqual(abs(biax), 1)
        self.assertAlmostEqual(abs(unit), 1)
        self.assertAlmostEqual(abs(move), 1)

    def test_wall(self):
        biax = random_quat_Xwall()
        self.assertAlmostEqual(abs(biax), 1)
        #print(biax.q)
        biax = random_quat_Ywall()
        self.assertAlmostEqual(abs(biax), 1)
        #print(biax.q)
        biax = random_quat_Zwall()
        self.assertAlmostEqual(abs(biax), 1)
        #print(biax.q)
        biax = random_move_Xwall(0.25)
        self.assertEqual(biax.q[2], 0)
        self.assertEqual(biax.q[3], 0)
        self.assertAlmostEqual(abs(biax), 1)
        #print(biax.q)
        biax = random_move_Ywall(0.25)
        self.assertEqual(biax.q[1], 0)
        self.assertEqual(biax.q[3], 0)
        self.assertAlmostEqual(abs(biax), 1)
        #print(biax.q)
        biax = random_move_Zwall(0.25)
        self.assertEqual(biax.q[1], 0)
        self.assertEqual(biax.q[2], 0)
        self.assertAlmostEqual(abs(biax), 1)
        #print(biax.q)

    def tearDown(self): pass

if __name__== "__main__":

    unittest.main()

# EOF



#!/usr/bin/env python3

try:
    real_types = (int, long, float)
    range = xrange
except NameError:   # Python 3
    real_types = (int, float)

import math
import random


class Quat:
    """The class defining a quaternion."""

    def __init__(self, x=0, y=0, z=0, t=0):
        """Create a Quat instance."""
        self.q = [float(x), float(y), float(z), float(t)]

    def __repr__(self):
        """Compute the string (formal) representation of the quaternion."""
        return "Quat({}, {}, {}, {})".format(*self.q)

    def __getitem__(self, key):
        """Implement quat[key]."""
        return self.q[key]

    def _normalize(self, other):
        """Transformation of an object to a quaternion."""
        if isinstance(other, real_types):
            other = Quat(other)
        elif isinstance(other, complex):
            other = Quat(other.real, other.imag)
        return other

    def __eq__(self, other):
        """Test if the quaternions are equal."""
        other = self._normalize(other)
        return self.q == other.q

    def __ne__(self, other):
        """Test if the quaternions are not equal."""
        return not self == other

    def __nonzero__(self):
        """Test if the quaternion is not equal to zero."""
        return self.q != [0.0, 0.0, 0.0, 0.0]

    __bool__ = __nonzero__   # Python 3

    def __pos__(self):
        """Implementation of +q."""
        return self

    def __neg__(self):
        """Implementation of -q."""
        return Quat(-self.q[0], -self.q[1], -self.q[2], -self.q[3])

    def __add__(self, other):
        """Addition of quaternions."""
        other = self._normalize(other)
        return Quat(self.q[0] + other.q[0], self.q[1] + other.q[1],
            self.q[2] + other.q[2], self.q[3] + other.q[3])

    __radd__ = __add__

    def __sub__(self, other):
        """Subtraction of quaternions."""
        # return self + (-other)
        other = self._normalize(other)
        return Quat(self.q[0] - other.q[0], self.q[1] - other.q[1],
            self.q[2] - other.q[2], self.q[3] - other.q[3])

    def __rsub__(self, other):
        """Subtraction of quaternions."""
        # return (-self) + other
        other = self._normalize(other)
        return Quat(other.q[0] - self.q[0], other.q[1] - self.q[1],
            other.q[2] - self.q[2], other.q[3] - self.q[3])

    def __mul__(self, other):
        """Quaternion product."""
        other = self._normalize(other)
        a = (self.q[0] * other.q[0] - self.q[1] * other.q[1]
            - self.q[2] * other.q[2] - self.q[3] * other.q[3])
        b = (self.q[0] * other.q[1] + self.q[1] * other.q[0]
            + self.q[2] * other.q[3] - self.q[3] * other.q[2])
        c = (self.q[0] * other.q[2] - self.q[1] * other.q[3]
            + self.q[2] * other.q[0] + self.q[3] * other.q[1])
        d = (self.q[0] * other.q[3] + self.q[1] * other.q[2]
            - self.q[2] * other.q[1] + self.q[3] * other.q[0])
        return Quat(a, b, c, d)

    def __rmul__(self, other):
        """Quaternion product."""
        other = self._normalize(other)
        a = (other.q[0] * self.q[0] - other.q[1] * self.q[1]
            - other.q[2] * self.q[2] - other.q[3] * self.q[3])
        b = (other.q[0] * self.q[1] + other.q[1] * self.q[0]
            + other.q[2] * self.q[3] - other.q[3] * self.q[2])
        c = (other.q[0] * self.q[2] - other.q[1] * self.q[3]
            + other.q[2] * self.q[0] + other.q[3] * self.q[1])
        d = (other.q[0] * self.q[3] + other.q[1] * self.q[2]
            - other.q[2] * self.q[1] + other.q[3] * self.q[0])
        return Quat(a, b, c, d)

    def __abs__(self):
        """Return the norm of a quaternion (a scalar)."""
        powers = sum(item * item for item in self.q)
        return math.sqrt(powers)

    norm = __abs__

    def norm_squared(self):
        """Return the squared norm of a quaternion (a scalar)."""
        return sum(item * item for item in self.q)

    def normalized(self):
        """Return a normalized version of this quaternion."""
        a = abs(self)
        return Quat(self.q[0] / a, self.q[1] / a, self.q[2] / a, self.q[3] / a)

    def is_unit(self):
        """Test a unit quaternion."""
        return 1.0 == sum(item * item for item in self.q)

    def conjugate(self):
        """Conjugate the quaternion."""
        return Quat(self.q[0], -self.q[1], -self.q[2], -self.q[3])

    def __invert__(self):   # ~p, return p^{-1}
        """Reciprocal of the quaternion."""
        powers = sum(item * item for item in self.q)
        return (1.0 / powers) * self.conjugate()

    def _pow1(self, n):
        """Find powers of the quaternion (inefficient)."""
        if n < 0:
            return pow(~self, -n)
        quat = Quat(1)
        while n > 0:
            quat = quat * self
            n = n - 1
        return quat

    def _pow2(self, n):
        """Find powers of the quaternion (binary exponentiation)."""
        if n == 0:
            return Quat(1)
        if n < 0:
            return pow(~self, -n)
        quat = self
        if n == 1:
            return self
        elif n == 2:
            return self * self
        else:   # binary exponentiation
            result = Quat(1)
            while True:
                if n % 2 == 1:
                    result = result * quat
                    n = n - 1  # przez ile pomnozyc
                    if n == 0:
                        break
                if n % 2 == 0:
                    quat = quat * quat
                    n = n // 2
        return result

    __pow__ = _pow2

    def __hash__(self):
        """Hashable quaternions."""
        return hash(tuple(self.q))

    def __int__(self):
        """Conversion to int is not possible."""
        raise TypeError("can't convert quat to int")

    def __long__(self):
        """Conversion to long is not possible."""
        raise TypeError("can't convert quat to long")

    def __float__(self):
        """Conversion to float is not possible."""
        raise TypeError("can't convert quat to float")

    def __complex__(self):
        """Conversion to complex is not possible."""
        raise TypeError("can't convert quat to complex")

    def apply_to_vector(self, vector):
        """Return the rotated vector."""
        if self.is_unit():
            unit_quat = self
        else:
            print("not a unit quaternion")
            unit_quat = self.normalized()
        if len(vector) != 3:
            raise ValueError("not a 3D vector")
        vec_quat = Quat(0, vector[0], vector[1], vector[2])
        vec_quat = unit_quat * vec_quat * (~unit_quat)
        # zwracamy wektor z R^3 (wycinek)
        return vec_quat.q[1:]   # list or numpy.array

    def get_rotation_matrix(self):
        """Return the rotation matrix."""
        w, x, y, z = self.q

        w2 = w * w
        x2 = x * x
        y2 = y * y
        z2 = z * z
        xy = x * y
        wz = w * z
        xz = x * z
        wy = w * y
        yz = y * z
        wx = w * x

        inverse = 1.0 / (w2 + x2 + y2 + z2)
        inverse2 = 2.0 * inverse

        m00 = (w2 + x2 - y2 - z2) * inverse
        m01 = (xy - wz) * inverse2
        m02 = (xz + wy) * inverse2

        m10 = (xy + wz) * inverse2
        m11 = (w2 - x2 + y2 - z2) * inverse
        m12 = (yz - wx) * inverse2

        m20 = (xz - wy) * inverse2
        m21 = (yz + wx) * inverse2
        m22 = (w2 - x2 - y2 + z2) * inverse

        return [[m00, m01, m02, 0.],
                [m10, m11, m12, 0.],
                [m20, m21, m22, 0.],
                [0., 0., 0., 1.]]

    # method used to create a rotation Quaternion to rotate
    # any vector defined as a Quaternion
    # with respect to the vector vect theta 'radians';
    # rot_vec ma dlugosc 1 w R^3, tworzymy kwaternion jednostkowy.
    # Chyba lepiej zrobic metode klasy.
    @classmethod
    def rot_quat(cls, axis, angle):
        """From the axis-angle representation to the quat.
        The angle is in radians. The axis is a unit 3D vector."""
        if len(axis) != 3:
            raise ValueError("not a 3D vector")
        length = math.sqrt(sum(x * x for x in axis))
        if length != 1.0:
            print("not a unit vector")
            axis = [x / length for x in axis]
        q0 = math.cos(angle / 2.0)
        sinus = math.sin(angle / 2.0)
        q1 = axis[0] * sinus
        q2 = axis[1] * sinus
        q3 = axis[2] * sinus
        return cls(q0, q1, q2, q3)

    @classmethod
    def from_x_rotation(cls, angle):
        """Create the unit quat for the X rotation."""
        return cls.rot_quat([1, 0, 0], angle)

    @classmethod
    def from_y_rotation(cls, angle):
        """Create the unit quat for the Y rotation."""
        return cls.rot_quat([0, 1, 0], angle)

    @classmethod
    def from_z_rotation(cls, angle):
        """Create the unit quat for the Z rotation."""
        return cls.rot_quat([0, 0, 1], angle)

    create_from_axis_rotation = rot_quat
    create_from_x_rotation = from_x_rotation
    create_from_y_rotation = from_y_rotation
    create_from_z_rotation = from_z_rotation

    @classmethod
    def from_eulers(cls, phi, theta, psi):
        """Create the unit quat from Euler angles."""
        unit_quat = cls.from_z_rotation(phi)
        unit_quat *= cls.from_y_rotation(theta)
        unit_quat *= cls.from_z_rotation(psi)
        return unit_quat

    @classmethod
    def random_quat_uniax(cls):
        """Return a random rotation quat for uniaxial molecules."""
        phi = random.uniform(0, 2*math.pi)
        ct = random.uniform(-1, 1)
        theta = math.acos(ct)   # mozna tez -ct
        quat = cls.from_eulers(phi, theta, 0)
        return quat

    @classmethod
    def random_quat_biax(cls):
        """Return a random rotation quat for biaxial molecules."""
        phi = random.uniform(0, 2*math.pi)
        ct = random.uniform(-1, 1)
        theta = math.acos(ct)   # mozna tez -ct
        psi = random.uniform(0, 2*math.pi)
        quat = cls.from_eulers(phi, theta, psi)
        return quat

    @staticmethod
    def random_pair():
        """Return a random pair (s1, s2), where s1^2 + s2^2 < 1."""
        while True:
            s1 = random.uniform(-1, 1)
            s2 = random.uniform(-1, 1)
            if s1 * s1 + s2 * s2 < 1:
                break
        return s1, s2

    @classmethod
    def random_unit_quat(cls):
        """Return a random unit quat (Marsaglia, 1972)."""
        s1, s2 = cls.random_pair()
        s3, s4 = cls.random_pair()
        S1 = s1 * s1 + s2 * s2
        S2 = s3 * s3 + s4 * s4
        a = math.sqrt((1.0 - S1) / S2)
        return cls(s1, s2, s3 * a, s4 * a)

    @classmethod
    def random_move_quat(cls, ksi=1.0):
        """Return a Monte Carlo move (Barker, Watts, 1969)."""
        assert 0.0 <= ksi <= 1.0
        s1, s2 = cls.random_pair()
        S = s1 * s1 + s2 * s2
        q1 = 2 * s1 * math.sqrt(1-S)
        q2 = 2 * s2 * math.sqrt(1-S)
        q3 = 1 - 2 * S
        # [q1, q2, q3] is now a unit vector.
        angle = random.uniform(-1, 1) * math.pi * 0.5 * 0.5 * ksi
        # math.pi * 0.5 gives all rotations,
        # 0.5 is for quat cos(angle/2), sin(angle/2).
        q0 = math.cos(angle)
        sinus = math.sin(angle)
        q1 *= sinus
        q2 *= sinus
        q3 *= sinus
        return cls(q0, q1, q2, q3)

Quaternion = Quat

# EOF



#!/usr/bin/env python3
#
# Przyklad jednoczesnego sortowania wartosci wlasnych i wektorow wlasnych.
# The eigenvalues are not necessarily ordered.
# help(numpy.argsort)
# argsort(a, axis=-1, kind='quicksort', order=None)
# Returns the indices that would sort an array.
#
# https://stackoverflow.com/questions/33037616/large-matrix-diagonalization-python

import math
import numpy

A = numpy.array([[3,1,-1], [1,3,-1], [-1,-1,5]], dtype=float)
print ( A )
# Wartosci wlasne to 2, 3, 6, ale kolejnosc moze byc rozna.
# Wektory wlasne beda ulozone kolumnami!
# The column vec[:,i] is the eigenvector corresponding to the eigenvalue w[i].
w, vec =  numpy.linalg.eig(A)
#print (w, vec)
print ("eigenvalues", w)
print ("unit eigenvectors (in columns)")
print (vec)

# Zmieniam kolejnosc wartosci wlasnych i wektorow wlasnych.
print ( "Zmiana kolejnosci ..." )
idx = w.argsort()[::-1] #large to small
# idx = w.argsort() #small to large
# idx to jest index_array : ndarray, int.
w = w[idx]   # to dziala elementwise, faktycznie sortuje w
vec = vec[:,idx]   # wiersze po kolei, kolumny wg idx
print (w, vec)
# now they are ordered and you can iterate through your results
# to write them to your file

print ( "Sprawdzenie dzialania macierzy na wektory ..." )
# Mnozenie macierzy i wektora wlasnego ma byc rowne mnozeniu
# wartosci wlasnej i wektora wlasnego.
v0 = vec[:,0] # kolejne wektory wlasne z kolumn
v1 = vec[:,1]
v2 = vec[:,2]
print ( "A v", numpy.dot(A,v0) )
print ( "lam v", w[0] * v0 )
print ( "A v", numpy.dot(A,v1) )
print ( "lam v", w[1] * v1 )
print ( "A v", numpy.dot(A,v2) )
print ( "lam v", w[2] * v2 )   # bledy maszynowe

print ( "Diagonalizacja B^T A B ..." )
print ( numpy.dot(numpy.transpose(vec), numpy.dot(A,vec)) ) # dziala!

# EOF
